import tempfile
import shutil
from pathlib import Path
from io import BytesIO
import click
from pyspark.sql import SparkSession, DataFrameWriter
import carduus.token as t
from carduus.keys import _PRIVATE_KEY_ENV_VAR, _RECIPIENT_PUBLIC_KEY_ENV_VAR


# All OPPRL token specifications in a dictionary for easy lookup by the token name.
TOKEN_SPECS: dict[str, t.OpprlToken] = {
    "opprl_token_1": t.OpprlToken.token1,
    "opprl_token_2": t.OpprlToken.token2,
    "opprl_token_3": t.OpprlToken.token3,
}


# The PII fields that each OPPRL token depends on.
TOKEN_PII_DEPENDENCIES: dict[str, list[t.OpprlPii]] = {
    "opprl_token_1": [
        t.OpprlPii.first_name,
        t.OpprlPii.last_name,
        t.OpprlPii.gender,
        t.OpprlPii.birth_date,
    ],
    "opprl_token_2": [
        t.OpprlPii.first_name,
        t.OpprlPii.last_name,
        t.OpprlPii.gender,
        t.OpprlPii.birth_date,
    ],
    "opprl_token_3": [
        t.OpprlPii.first_name,
        t.OpprlPii.last_name,
        t.OpprlPii.gender,
        t.OpprlPii.birth_date,
    ],
}


def pii_transforms_for_tokens(tokens: list[str]) -> dict[str, t.PiiTransform]:
    pii_transforms = {}
    for token in tokens:
        for pii_field in TOKEN_PII_DEPENDENCIES[token]:
            pii_transforms[pii_field.name] = pii_field.value
    return pii_transforms


def get_spark(num_threads: int | None) -> SparkSession:
    """Gets active spark session or creates a new local spark session."""
    p = str(num_threads) if num_threads else "*"
    return (
        SparkSession.Builder().master(f"local[{p}]").appName("Spindle Token CLI").getOrCreate()
    )


def write_single_file(dfw: DataFrameWriter, path: str):
    """Writes a 1 partition PySpark Dataframe to a single file."""
    tmp_dir = tempfile.mkdtemp(prefix="spindle-token-")
    try:
        tmp_path = Path(tmp_dir) / "data"
        dfw.save(str(tmp_path))
        part_file = next(tmp_path.rglob("part-*"), None)
        if part_file is None:
            raise FileNotFoundError("No data files found in the temporary directory.")
        shutil.copy(part_file, path)
    finally:
        shutil.rmtree(tmp_dir)


##################################################################
# CLI


def common_options(func):
    """A decorator that adds the click options that are found on all commands."""
    func = click.option(
        "-p",
        "--parallelism",
        envvar="SPINDLE_TOKEN_PARALLELISM",
        type=click.INT,
        help="The number of worker threads to parallelize over. Useful when the input dataset is "
        "partitioned into multiple part files. If not supplied, defaults to the number of logical cores.",
    )(func)
    func = click.option(
        "-f",
        "--format",
        envvar="SPINDLE_TOKEN_FORMAT",
        type=click.Choice(("parquet", "csv"), case_sensitive=False),
        required=True,
        help="The file format of input and output data files.",
    )(func)
    func = click.option(
        "-k",
        "--key",
        envvar=_PRIVATE_KEY_ENV_VAR + "_FILE",
        type=click.File(mode="rb"),
        help="The PEM file containing your private key.",
    )(func)
    return func


@click.group()
def cli():
    """A command line tool for tokenizing and transcrypting data files using the Open Privacy Preserving Record Linkage (OPPRL) protocol."""
    pass


@cli.command()
@click.option(
    "-t",
    "--token",
    type=click.Choice(list(TOKEN_SPECS.keys())),
    multiple=True,
    required=True,
    help="An OPPRL token to add to the dataset. Can be passed multiple times.",
)
@common_options
@click.argument("input", type=click.Path(exists=True))
@click.argument("output", type=click.Path(exists=False))
def tokenize(
    input: str,
    output: str,
    key: BytesIO | None,
    token: list[str],
    format: str,
    parallelism: int | None,
):
    """Add tokens to a dataset of PII.

    Creates a dataset at the OUTPUT location that adds encrypted OPPRL tokens to the INPUT dataset.
    Does not modify the INPUT dataset.

    INPUT is the path to the dataset to tokenize. If INPUT is a file, it must be of the format provided to
    the `--format` option. If INPUT is a directory, all files within the directory that match the given
    format will be considered a partition of the dataset.

    OUTPUT is the file or directory in which the tokenized dataset will be written. If INPUT is a
    file, the OUTPUT will be written to as a file. If INPUT is a directory, the OUTPUT will be a directory
    containing a dataset partitioned into files.
    """
    input_path = Path(input)
    output_path = Path(output)
    tokens = [TOKEN_SPECS[t] for t in token]
    tokens.sort(key=lambda t: t.name)
    pii_transforms = pii_transforms_for_tokens(token)

    spark = get_spark(parallelism)
    df = spark.read.format(format).option("delimiter", "|").option("header", True).load(input)

    df = t.tokenize(
        df,
        pii_transforms=pii_transforms,
        tokens=tokens,
        private_key=key.read() if key else None,
    )

    if input_path.is_file():
        df = df.repartition(1)

    dfw = df.write.format(format).option("delimiter", "|").option("header", True)
    if input_path.is_file():
        write_single_file(dfw, str(output_path))
    else:
        dfw.save(str(output_path))


@cli.group()
def transcrypt():
    """Prepare tokenized datasets to be sent or received."""
    pass


@transcrypt.command()
@click.option(
    "-t",
    "--token",
    type=click.Choice(list(TOKEN_SPECS.keys())),
    multiple=True,
    required=True,
    help="The column name of an OPPRL token on the input data to transcrypt.",
)
@click.option(
    "-r",
    "--recipient",
    envvar=_RECIPIENT_PUBLIC_KEY_ENV_VAR + "_FILE",
    type=click.File(mode="rb"),
    help="The PEM file containing the recipients public key.",
)
@common_options
@click.argument("input", type=click.Path(exists=True))
@click.argument("output", type=click.Path(exists=False))
def out(
    input: str,
    output: str,
    key: BytesIO | None,
    recipient: BytesIO | None,
    token: list[str],
    format: str,
    parallelism: int | None,
):
    """Prepare ephemeral tokens for a specific recipient.

    INPUT is the path to the dataset of tokens to create ephemeral tokens from. If INPUT is a file, it must
    be of the format provided to the `--format` option. If INPUT is a directory, all files within the
    directory that match the given format will be considered a partition of the dataset.

    OUTPUT is the file or directory in which the tokenized dataset will be written. If INPUT is a
    file, the OUTPUT will be written to as a file. If INPUT is a directory, the OUTPUT will be a directory
    containing a dataset partitioned into files.
    """
    input_path = Path(input)
    output_path = Path(output)
    spark = get_spark(parallelism)
    df = spark.read.format(format).option("delimiter", "|").option("header", True).load(input)

    df = t.transcrypt_out(
        df,
        token_columns=token,
        recipient_public_key=recipient.read() if recipient else None,
        private_key=key.read() if key else None,
    )

    if input_path.is_file():
        df = df.repartition(1)

    dfw = df.write.format(format).option("delimiter", "|").option("header", True)
    if input_path.is_file():
        write_single_file(dfw, str(output_path))
    else:
        dfw.save(str(output_path))


@transcrypt.command("in")
@click.option(
    "-t",
    "--token",
    type=click.Choice(list(TOKEN_SPECS.keys())),
    multiple=True,
    required=True,
    help="The column name of an OPPRL token on the input data to transcrypt.",
)
@common_options
@click.argument("input", type=click.Path(exists=True))
@click.argument("output", type=click.Path(exists=False))
def in_(
    input: str,
    output: str,
    key: BytesIO | None,
    token: list[str],
    format: str,
    parallelism: int | None,
):
    """Convert a dataset of ephemeral tokens into tokens.

    INPUT is the path to the dataset of ephemeral tokens to create tokens from. If INPUT is a file, it must
    be of the format provided to the `--format` option. If INPUT is a directory, all files within the
    directory that match the given format will be run.

    OUTPUT is the file or directory in which the tokenized dataset will be written. If INPUT is a
    file, the OUTPUT will be written to as a file. If INPUT is a directory, the OUTPUT will be a directory
    containing a dataset partitioned into files.

    """
    input_path = Path(input)
    output_path = Path(output)
    spark = get_spark(parallelism)
    df = spark.read.format(format).option("delimiter", "|").option("header", True).load(input)

    df = t.transcrypt_in(
        df,
        token_columns=token,
        private_key=key.read() if key else None,
    )

    if input_path.is_file():
        df = df.repartition(1)

    dfw = df.write.format(format).option("delimiter", "|").option("header", True)
    if input_path.is_file():
        write_single_file(dfw, str(output_path))
    else:
        dfw.save(str(output_path))



import os
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.hashes import Hash, SHAKE256


def generate_pem_keys(key_size: int = 2048) -> tuple[bytes, bytes]:
    """Generates a fresh RSA key pair.

    Arguments:
        key_size:
            The size (in bits) of the key.

    Returns:
        A tuple containing the private key and public key bytes. Both in the PEM encoding.

    """
    key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=key_size,
    )
    private = key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption(),
    )
    public = key.public_key().public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo,
    )
    return private, public


def derive_aes_key(rsa_key: bytes) -> bytes:
    """Derives the corresponding AES key from the given RSA private key.

    Arguments:
        rsa_key:
            A RSA private key.

    Returns:
        An 32 byte AES key.

    """
    digest = Hash(SHAKE256(32))
    digest.update(rsa_key)
    return digest.finalize()


_PRIVATE_KEY_ENV_VAR = "SPINDLE_TOKEN_PRIVATE_KEY"
_RECIPIENT_PUBLIC_KEY_ENV_VAR = "SPINDLE_TOKEN_RECIPIENT_PUBLIC_KEY"


def private_key_from_env() -> bytes:
    if _PRIVATE_KEY_ENV_VAR not in os.environ:
        raise ValueError(
            f"No private RSA key found. Set the {_PRIVATE_KEY_ENV_VAR} environment variable or pass the key as an argument."
        )
    return os.environ[_PRIVATE_KEY_ENV_VAR].encode()


def public_key_from_env() -> bytes:
    if _RECIPIENT_PUBLIC_KEY_ENV_VAR not in os.environ:
        raise ValueError(
            f"No public RSA key set. Set the {_RECIPIENT_PUBLIC_KEY_ENV_VAR} environment variable or pass the key as an argument."
        )
    return os.environ[_RECIPIENT_PUBLIC_KEY_ENV_VAR].encode()



from abc import ABC
from dataclasses import dataclass

from pyspark.sql import Column, DataFrame
from pyspark.sql.functions import (
    array,
    array_join,
    col,
    when,
    year,
    month,
    dayofmonth,
    coalesce,
    soundex,
    regexp_replace,
    to_date,
)
from pyspark.sql.types import (
    DataType,
    StringType,
    DateType,
    TimestampType,
    TimestampNTZType,
)

from carduus.token._impl import null_safe, remap, normalize_text, first_char, metaphone


class PiiTransform(ABC):
    """Abstract base class for normalization and enhancement of a specific PII attribute.

    Intended to be extended by users to add support for building tokens from a custom PII attribute.
    """

    def normalize(self, column: Column, dtype: DataType) -> Column:
        """A normalized representation of the PII column.

        A normalized value has eliminated all representation or encoding differences so all instances of the
        same logical values have identical physical values. For example, text attributes will often be normalized
        by filtering to alpha-numeric characters and whitespace, standardizing to whitespace to the space character,
        and converting all alpha characters to uppercase to ensure that all ways of representing the same phrase
        normalize to the exact same string.

        Arguments:
            column:
                The spark `Column` expression for the PII attribute being normalized.
            dtype:
                The spark `DataType` object of the `column` object found on the `DataFrame` being normalized.
                Can be used to delegate to different normalization logic based on different schemas of input data.
                For example, a subject's birth date may be a `DateType`, `StringType`, or `LongType` on input data and
                thus requires corresponding normalization into a `DateType`.

        Returns:
            The normalized version of the PII attribute.
        """
        return column

    def enhancements(self, column: Column) -> dict[str, Column]:
        """A collection of PII attributes that can be automatically derived from a given normalized PII attribute


        If an implementation of [PiiTransform][carduus.token.PiiTransform] does not override this method, it is
        assumed that no enhancements can be derived

        Arguments:
            column:
                The normalized PII column to produce enhancements from.

        Return:
            A `dict` with keys that correspond to the PII attributes of the ___ and values that correspond to the `Column`
            expression that produced the new PII from a normalized input attribute.

        """
        return {}


@dataclass
class NameTransform(PiiTransform):
    enhancement_prefix: str

    def normalize(self, column: Column, _: DataType) -> Column:
        # @TODO Should spaces be stripped out?
        return normalize_text(regexp_replace(column, "[^a-zA-Z ]", ""))

    def enhancements(self, column: Column) -> dict[str, Column]:
        return {
            self.enhancement_prefix + "_initial": first_char(column),
            self.enhancement_prefix + "_soundex": soundex(column),
            self.enhancement_prefix + "_metaphone": metaphone(column),
        }


class GenderTransform(PiiTransform):

    _lookup = {
        "F": "F",
        "W": "F",
        "G": "F",
        "M": "M",
        "B": "M",
    }

    def normalize(self, column: Column, _: DataType) -> Column:
        return coalesce(
            remap(self._lookup, first_char(normalize_text(column))),
            when(column.isNotNull(), "O"),
        )


@dataclass
class DateTransform(PiiTransform):
    enhancement_prefix: str
    date_format: str = "yyyy-MM-dd"

    def normalize(self, column: Column, from_type: DataType) -> Column:
        if isinstance(from_type, (DateType, TimestampType, TimestampNTZType)):
            return column
        if isinstance(from_type, StringType):
            return to_date(column, self.date_format)
        raise Exception(
            f"Cannot normalize column of type {from_type} into a DateType column. Column: {column}."
        )

    def enhancements(self, column: Column) -> dict[str, Column]:
        return {
            self.enhancement_prefix + "_year": year(column),
            self.enhancement_prefix + "_month": month(column),
            self.enhancement_prefix + "_day": dayofmonth(column),
        }


@dataclass
class Pii:
    data: DataFrame
    pii_columns: set[str]


def normalize_pii(
    df: DataFrame,
    pii_transforms: dict[str, PiiTransform],
) -> DataFrame:
    return df.select(
        *[
            (
                pii_transforms[column]
                .normalize(df[column], df.schema[column].dataType)
                .alias(column)
                if column in pii_transforms
                else col(column)
            )
            for column in df.columns
        ]
    )


def enhance_pii(
    df: DataFrame,
    pii_transforms: dict[str, PiiTransform],
) -> tuple[DataFrame, set[str]]:
    new_pii: set[str] = set()
    all_columns = []
    for column in df.columns:
        all_columns.append(col(column))
        if column in pii_transforms:
            enhancements = pii_transforms[column].enhancements(col(column))
            all_columns.extend(
                [
                    enhancement_col.alias(enhancement_name)
                    for enhancement_name, enhancement_col in enhancements.items()
                    if enhancement_name not in df.columns
                ]
            )
            new_pii = new_pii | set(enhancements.keys())
    return df.select(*all_columns), new_pii


@null_safe
def join_pii(*args: Column):
    return array_join(array(*args), delimiter=":")



from typing import Any, Callable
import jellyfish as j

from pyspark.sql import Column
from pyspark.sql.functions import (
    lit,
    when,
    length,
    substring,
    upper,
    regexp_replace,
    trim,
    udf,
    base64,
    replace,
)
from pyspark.sql.types import StringType


def remap(mapping: dict[Any, Any], col: Column, default: Any | None = None) -> Column:
    result = None
    for k, v in mapping.items():
        if result is None:
            result = when(col == lit(k), lit(v))
        else:
            result = result.when(col == lit(k), lit(v))
    if default is not None and result is not None:
        result = result.otherwise(default)
    return col if result is None else result


def null_if(pred: Callable[[Column], Column], col: Column) -> Column:
    return when(~pred(col), col).otherwise(lit(None))


def first_char(col: Column) -> Column:
    return substring(col, 1, 1)


def _metaphone_udf_impl(s: str | None) -> str | None:
    return j.metaphone(s) if s else None


metaphone = udf(_metaphone_udf_impl, returnType=StringType(), useArrow=False)


def normalize_text(raw: Column) -> Column:
    return null_if(lambda c: length(c) == 0, trim(regexp_replace(upper(raw), "\\s+", " ")))


def null_safe(func):
    """A decorator for functions that operate over pyspark Column objects.
    If any of the input columns are NULL, the result will be NULL. Otherwise the decorated function will be called.
    """

    def wrapper(*args: Column) -> Column:
        if len(args) == 0:
            return func()
        result: Column = when(args[0].isNull(), lit(None))
        for arg in args[1:]:
            result = result.when(arg.isNull(), lit(None))
        return result.otherwise(func(*args))

    return wrapper


def base64_no_newline(col: Column) -> Column:
    """The default behavior of Spark's base64 function adds newlines every 76 characters for parity with unix.
    Although this behavior doesn't make tokens invalid, it is likely to cause issues with users working with CSV files
    that contain tokens.

    See: https://superuser.com/questions/1225134/why-does-the-base64-of-a-string-contain-n
    """
    return replace(base64(col), lit("\r\n"), lit(""))



import os
from collections.abc import Iterable, Sequence, Mapping
from dataclasses import dataclass
from enum import Enum
from pyspark.sql import DataFrame
from pyspark.sql.functions import (
    col,
    lit,
    sha2,
    udf,
    to_binary,
)
from pyspark.sql.types import BinaryType
from cryptography.hazmat.primitives.serialization import (
    load_pem_private_key,
    load_pem_public_key,
)
from carduus.token.pii import (
    normalize_pii,
    enhance_pii,
    join_pii,
    PiiTransform,
    NameTransform,
    GenderTransform,
    DateTransform,
)
import carduus.token.crypto as crypto
from carduus.token._impl import base64_no_newline
from carduus.keys import derive_aes_key, private_key_from_env, public_key_from_env


__all__ = [
    "tokenize",
    "transcrypt_out",
    "transcrypt_in",
    "OpprlPii",
    "OpprlToken",
    "PiiTransform",
    "TokenSpec",
]


class OpprlPii(Enum):
    """Enum of [PiiTransform][carduus.token.PiiTransform] objects for the PII fields supported by the Open Privacy Preserving
    Record Linkage specification.

    Attributes:
        first_name:
            [PiiTransform][carduus.token.PiiTransform] implementation for a subject's first name according to the OPPRL standard.
        middle_name:
            [PiiTransform][carduus.token.PiiTransform] implementation for a subject's middle name according to the OPPRL standard.
        last_name:
            [PiiTransform][carduus.token.PiiTransform] implementation for a subject's last (aka family) name according to the OPPRL standard.
        gender:
            [PiiTransform][carduus.token.PiiTransform] implementation for a subject's gender according to the OPPRL standard.
        birth_date:
            [PiiTransform][carduus.token.PiiTransform] implementation for a subject's date of birth according to the OPPRL standard.

    """

    first_name: NameTransform = NameTransform("first")
    middle_name: NameTransform = NameTransform("middle")
    last_name: NameTransform = NameTransform("last")
    gender: GenderTransform = GenderTransform()
    birth_date: DateTransform = DateTransform("birth")  # @TODO: Parameterize date format


@dataclass(frozen=True)
class TokenSpec:
    """An collection of PII fields that will be encrypted together to create a token.

    For an enum of standard `TokenSpec` instances that comply with the Open Privacy Preserving Record Linkage protocol
    see [`OpprlToken`][carduus.token.OpprlToken].

    Attributes:
        name:
            The name of the column that holds these tokens.
        fields:
            The PII fields to encrypt together to create token values.
    """

    name: str
    fields: Iterable[str]


class OpprlToken(Enum):
    """Enum of [`TokenSpec`][carduus.token.TokenSpec] objects that meet the Open Privacy Preserving
    Record Linkage tokenization specification.

    Attributes:
        opprl_token_1:
            Token #1 from the OPPRL specification. Creates tokens based on `first_initial`, `last_name`, `gender`, and `birth_date`.
        opprl_token_2:
            Token #2 from the OPPRL specification. Creates tokens based on `first_soundex`, `last_soundex`, `gender`, and `birth_date`.
        orrpl_token_3:
            Token #3 from the OPPRL specification. Creates tokens based on the `first_metaphone`, `last_metaphone`, `gender`, and `birth_date`.

    """

    token1: TokenSpec = TokenSpec(
        "opprl_token_1",
        (
            "first_initial",
            OpprlPii.last_name.name,
            OpprlPii.gender.name,
            OpprlPii.birth_date.name,
        ),
    )

    token2: TokenSpec = TokenSpec(
        "opprl_token_2",
        (
            "first_soundex",
            "last_soundex",
            OpprlPii.gender.name,
            OpprlPii.birth_date.name,
        ),
    )

    token3: TokenSpec = TokenSpec(
        "opprl_token_3",
        (
            "first_metaphone",
            "last_metaphone",
            OpprlPii.gender.name,
            OpprlPii.birth_date.name,
        ),
    )


def tokenize(
    df: DataFrame,
    pii_transforms: Mapping[str, PiiTransform | OpprlPii],
    tokens: Sequence[TokenSpec | OpprlToken],
    private_key: bytes | None = None,
) -> DataFrame:
    """Adds encrypted token columns based on PII.

    All PII columns found in the `DataFrame` are normalized using the provided `pii_transforms`.
    All PII attributes provided by the enhancements of the `pii_transforms` are added if they
    are not already present in the `DataFrame`. The fields of each [`TokenSpec`][carduus.token.TokenSpec]
    from `tokens` are hashed and encrypted together according to the OPPRL specification.

    Arguments:
        df:
            The pyspark `DataFrame` containing all PII attributes.
        pii_transforms:
            A dictionary that maps column names of `df` to [PiiTransform][carduus.token.PiiTransform]
            objects to specify how each raw PII column is normalized and enhanced into derived PII attributes.
            Values can also be a member of the [OpprlPii][carduus.token.OpprlPii] enum if using the
            standard OPPRL tokens.
        tokens:
            A collection of [`TokenSpec`][carduus.token.TokenSpec] objects that denotes which PII attributes
            are encrypted into each token. Elements can also be a member of the [OpprlToken][carduus.token.OpprlToken]
            enum if using the standard OPPRL tokens.
        private_key:
            Your private RSA key. This argument should only be set when reading from a secrets manager or testing, otherwise it is
            recommended to set the SPINDLE_TOKEN_PRIVATE_KEY environment variable with your private key.

    Returns:
        The `DataFrame` with PII columns replaced by encrypted tokens.

    """
    if not private_key:
        private_key = private_key_from_env()

    # Raise clear error message if key is invalid.
    load_pem_private_key(private_key, None)
    pii_transforms_ = {
        c: tr.value if isinstance(tr, OpprlPii) else tr for c, tr in pii_transforms.items()
    }
    tokens_ = [t.value if isinstance(t, OpprlToken) else t for t in tokens]
    if len(tokens) == 0:
        return df
    token_columns = [token.name for token in tokens_]
    encrypt = udf(
        crypto.make_deterministic_encrypter(derive_aes_key(private_key)),
        returnType=BinaryType(),
    )

    pii = normalize_pii(df, pii_transforms_)
    pii, new_pii_columns = enhance_pii(pii, pii_transforms_)

    return (
        pii.withColumns(
            {t.name: join_pii(*[col(f) for f in sorted(t.fields)]) for t in tokens_}
        )
        .withColumns(
            {
                column: base64_no_newline(
                    encrypt(to_binary(sha2(col(column), 512), lit("hex")))
                )
                for column in token_columns
            }
        )
        .drop(*new_pii_columns)
    )


def transcrypt_out(
    df: DataFrame,
    token_columns: Iterable[str],
    recipient_public_key: bytes | None = None,
    private_key: bytes | None = None,
) -> DataFrame:
    """Prepares a `DataFrame` containing encrypted tokens to be sent to a specific trusted party by re-encrypting
    the tokens using the recipient's public key without exposing the original PII.

    Output tokens will be unmatchable to any dataset or within the given dataset until the intended recipient
    processes the data with [`transcrypt_in`][carduus.token.transcrypt_in].

    Arguments:
        df:
            Spark `DataFrame` with token columns to transcrypt.
        token_columns:
            The collection of column names that correspond to tokens.
        recipient_public_key:
            The public RSA key of the recipient who will be receiving the dataset with ephemeral tokens. Can also be supplied
            the SPINDLE_TOKEN_RECIPIENT_PUBLIC_KEY environment variable.
        private_key:
            Your private RSA key. This argument should only be set when reading from a secrets manager or testing, otherwise it is
            recommended to set the SPINDLE_TOKEN_PRIVATE_KEY environment variable with your private key.

    Returns:
        The `DataFrame` with the original encrypted tokens re-encrypted for sending to the recipient.
    """
    if not recipient_public_key:
        recipient_public_key = public_key_from_env()
    if not private_key:
        private_key = private_key_from_env()

    # Raise clear error message if key is invalid.
    load_pem_private_key(private_key, None)
    load_pem_public_key(recipient_public_key)
    decrypt = udf(
        crypto.make_deterministic_decrypter(derive_aes_key(private_key)),
        returnType=BinaryType(),
    )
    encrypt = udf(
        crypto.make_asymmetric_encrypter(recipient_public_key),
        returnType=BinaryType(),
    )
    return df.withColumns(
        {
            column: base64_no_newline(encrypt(decrypt(to_binary(col(column), lit("base64")))))
            for column in token_columns
        }
    )


def transcrypt_in(
    df: DataFrame,
    token_columns: Iterable[str],
    private_key: bytes | None = None,
) -> DataFrame:
    """Used by the recipient of a `DataFrame` containing tokens in the intermediate representation produced by
    [`transcrypt_out`][carduus.token.transcrypt_out] to re-encrypt the tokens such that they will match with
    other datasets

    Arguments:
        df:
            Spark `DataFrame` with token columns to transcrypt.
        token_columns:
            The collection of column names that correspond to tokens.
        private_key:
            Your private RSA key. This argument should only be set when reading from a secrets manager or testing, otherwise it is
            recommended to set the SPINDLE_TOKEN_PRIVATE_KEY environment variable with your private key.

    Returns:
        The `DataFrame` with the original encrypted tokens re-encrypted for sending to the destination.

    """
    if not private_key:
        private_key = private_key_from_env()

    # Raise clear error message if key is invalid.
    load_pem_private_key(private_key, None)
    decrypt = udf(crypto.make_asymmetric_decrypter(private_key), returnType=BinaryType())
    encrypt = udf(
        crypto.make_deterministic_encrypter(derive_aes_key(private_key)),
        returnType=BinaryType(),
    )
    return df.withColumns(
        {
            column: base64_no_newline(encrypt(decrypt(to_binary(col(column), lit("base64")))))
            for column in token_columns
        }
    )



from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding, rsa


# For tokenizing PII, We use the same nonce to encrypt
# The AES GCM SIV encryption scheme does NOT fail catastrophically when a nonce is
# reused, however an attacker can use a chosen-plaintext attack to attack specific
# records. https://en.wikipedia.org/wiki/Chosen-plaintext_attack
#
# When transferring data between parties we use RSA with OAEP so that datasets intercepted
# by untrusted third parties cannot exploit this vulnerability.

DEFAULT_NONCE = bytes([0] * 12)


def make_deterministic_encrypter(key: bytes):
    def encrypt(message: bytearray | None) -> bytes | None:
        if not message:
            return None
        from cryptography.hazmat.primitives.ciphers.aead import AESGCMSIV

        return AESGCMSIV(key).encrypt(DEFAULT_NONCE, bytes(message), None)

    return encrypt


def make_deterministic_decrypter(key: bytes):
    def decrypt(message: bytearray | None) -> bytes | None:
        if not message:
            return None
        from cryptography.hazmat.primitives.ciphers.aead import AESGCMSIV

        return AESGCMSIV(key).decrypt(DEFAULT_NONCE, bytes(message), None)

    return decrypt


def make_asymmetric_encrypter(public_key: bytes):
    def encrypt(message: bytearray | None) -> bytes | None:
        if not message:
            return None
        from cryptography.hazmat.primitives.serialization import load_pem_public_key

        key = load_pem_public_key(public_key)
        if not isinstance(key, rsa.RSAPublicKey):
            raise TypeError(f"Incorrect key type. Expected RSA public key, got {type(key)}.")
        return key.encrypt(
            bytes(message),
            padding.OAEP(
                mgf=padding.MGF1(algorithm=hashes.SHA256()),
                algorithm=hashes.SHA256(),
                label=None,
            ),
        )

    return encrypt


def make_asymmetric_decrypter(private_key: bytes):
    def decrypt(message: bytearray | None) -> bytes | None:
        if not message:
            return None
        from cryptography.hazmat.primitives.serialization import load_pem_private_key

        key = load_pem_private_key(private_key, None)
        if not isinstance(key, rsa.RSAPrivateKey):
            raise TypeError(f"Incorrect key type. Expected RSA private key, got {type(key)}.")
        return key.decrypt(
            bytes(message),
            padding.OAEP(
                mgf=padding.MGF1(algorithm=hashes.SHA256()),
                algorithm=hashes.SHA256(),
                label=None,
            ),
        )

    return decrypt
